@@ -6,9 +6,8 @@
 {-# LANGUAGE TemplateHaskell     #-}
 {-# LANGUAGE TupleSections       #-}
 
--- | Content addressable Haskell package management, providing for
--- secure, reproducible acquisition of Haskell package contents and
--- metadata.
+-- | Content addressable Haskell package management, providing for secure,
+-- reproducible acquisition of Haskell package contents and metadata.
 --
 -- @since 0.1.0.0
 module Pantry
@@ -22,6 +21,7 @@ module Pantry
   , defaultCasaRepoPrefix
   , defaultCasaMaxPerRequest
   , defaultSnapshotLocation
+  , defaultGlobalHintsLocation
   , HasPantryConfig (..)
   , withPantryConfig
   , withPantryConfig'
@@ -33,6 +33,7 @@ module Pantry
   , runPantryAppClean
   , runPantryAppWith
   , hpackExecutableL
+  , hpackForceL
 
     -- * Types
 
@@ -47,6 +48,9 @@ module Pantry
   , FlagName
   , PackageIdentifier (..)
 
+    -- ** Hpack types
+  , Hpack.Force (..)
+
     -- ** Files
   , FileSize (..)
   , RelFilePath (..)
@@ -106,6 +110,9 @@ module Pantry
   , SnapName (..)
   , snapshotLocation
 
+    -- ** Global hints
+  , GlobalHintsLocation (..)
+
     -- * Loading values
   , resolvePaths
   , loadPackageRaw
@@ -134,6 +141,7 @@ module Pantry
   , parseSnapName
   , parseRawSnapshotLocation
   , parsePackageIdentifierRevision
+  , parseRawPackageLocationImmutables
   , parseHackageText
 
     -- ** Cabal values
@@ -252,30 +260,32 @@ import           Pantry.Tree ( rawParseGPD, unpackTree )
 import           Pantry.Types as P
                    ( Archive (..), ArchiveLocation (..), BlobKey (..)
                    , CabalFileInfo (..), CabalString (..), FileSize (..)
-                   , FuzzyResults (..), HackageSecurityConfig (..)
-                   , HasPantryConfig (..), HpackExecutable (..), Mismatch (..)
-                   , ModuleName, Package (..), PackageCabal (..)
-                   , PackageIdentifier (..), PackageIdentifierRevision (..)
-                   , PackageIndexConfig (..), PackageLocation (..)
-                   , PackageLocationImmutable (..), PackageMetadata (..)
-                   , PackageName, PantryConfig (..), PantryException (..)
-                   , PHpack (..), PrintWarnings (..), RawArchive (..)
-                   , RawPackageLocation (..), RawPackageLocationImmutable (..)
-                   , RawPackageMetadata (..), RawSnapshot (..)
-                   , RawSnapshotLayer (..), RawSnapshotLocation (..)
-                   , RawSnapshotPackage (..), RelFilePath (..), Repo (..)
-                   , RepoType (..), ResolvedPath (..), Revision (..)
-                   , SafeFilePath, SHA256, SimpleRepo (..), SnapName (..)
-                   , Snapshot (..), SnapshotCacheHash (..), SnapshotLayer (..)
+                   , FuzzyResults (..), GlobalHintsLocation (..)
+                   , HackageSecurityConfig (..), HasPantryConfig (..)
+                   , HpackExecutable (..), Mismatch (..), ModuleName
+                   , Package (..), PackageCabal (..), PackageIdentifier (..)
+                   , PackageIdentifierRevision (..), PackageIndexConfig (..)
+                   , PackageLocation (..), PackageLocationImmutable (..)
+                   , PackageMetadata (..), PackageName, PantryConfig (..)
+                   , PantryException (..), PHpack (..), PrintWarnings (..)
+                   , RawArchive (..), RawPackageLocation (..)
+                   , RawPackageLocationImmutable (..), RawPackageMetadata (..)
+                   , RawSnapshot (..), RawSnapshotLayer (..)
+                   , RawSnapshotLocation (..), RawSnapshotPackage (..)
+                   , RelFilePath (..), Repo (..), RepoType (..)
+                   , ResolvedPath (..), Revision (..), SafeFilePath, SHA256
+                   , SimpleRepo (..), SnapName (..), Snapshot (..)
+                   , SnapshotCacheHash (..), SnapshotLayer (..)
                    , SnapshotLocation (..), SnapshotPackage (..), Tree (..)
                    , TreeEntry (..), TreeKey (..), Unresolved, Version
                    , WantedCompiler (..), bsToBlobKey, cabalFileName
-                   , defaultHackageSecurityConfig, defaultSnapshotLocation
-                   , flagNameString, getGlobalHintsFile, mkSafeFilePath
-                   , moduleNameString, packageIdentifierString
-                   , packageNameString, parseFlagName, parseHackageText
-                   , parsePackageIdentifier, parsePackageIdentifierRevision
-                   , parsePackageName, parsePackageNameThrowing
+                   , defaultGlobalHintsLocation, defaultHackageSecurityConfig
+                   , defaultSnapshotLocation, flagNameString, getGlobalHintsFile
+                   , globalHintsLocation, mkSafeFilePath, moduleNameString
+                   , packageIdentifierString, packageNameString, parseFlagName
+                   , parseHackageText, parsePackageIdentifier
+                   , parsePackageIdentifierRevision, parsePackageName
+                   , parsePackageNameThrowing, parseRawPackageLocationImmutables
                    , parseRawSnapshotLocation, parseSnapName, parseTreeM
                    , parseVersion, parseVersionThrowing, parseWantedCompiler
                    , pirForHash, resolvePaths, snapshotLocation
@@ -287,14 +297,17 @@ import           Path
                    ( Abs, Dir, File, Path, (</>), filename, parent, parseAbsDir
                    , parseRelFile, toFilePath
                    )
-import           Path.IO ( doesFileExist, listDir, resolveDir' )
+import           Path.IO ( copyFile, doesFileExist, listDir, resolveDir' )
 import           RIO
 import qualified RIO.ByteString as B
 import           RIO.Directory ( getAppUserDataDirectory )
 import qualified RIO.FilePath as FilePath
 import qualified RIO.List as List
 import qualified RIO.Map as Map
-import           RIO.PrettyPrint ( HasTerm (..) )
+import           RIO.PrettyPrint
+                   ( HasTerm (..), blankLine, flow, line, pretty, prettyDebugL
+                   , prettyError, prettyInfoL, string
+                   )
 import           RIO.PrettyPrint.StylesUpdate
                    ( HasStylesUpdate (..), StylesUpdate )
 import           RIO.Process
@@ -324,7 +337,8 @@ formatYamlParseError file e =
   <> displayException e
 
 -- | Create a new 'PantryConfig' with the given settings. For a version where
--- the use of Casa (content-addressable storage archive) is optional, see
+-- Hpack's approach to overwriting Cabal files is configurable and the use of
+-- Casa (content-addressable storage archive) is optional, see
 -- 'withPantryConfig''.
 --
 -- For something easier to use in simple cases, see 'runPantryApp'.
@@ -349,11 +363,13 @@ withPantryConfig ::
      -- ^ Max casa keys to pull per request.
   -> (SnapName -> RawSnapshotLocation)
      -- ^ The location of snapshot synonyms
+  -> (WantedCompiler -> GlobalHintsLocation)
+     -- ^ The location of global hints
   -> (PantryConfig -> RIO env a)
      -- ^ What to do with the config
   -> RIO env a
 withPantryConfig root pic he count pullURL maxPerRequest =
-  withPantryConfig' root pic he count (Just (pullURL, maxPerRequest))
+  withPantryConfig' root pic he Hpack.NoForce count (Just (pullURL, maxPerRequest))
 
 -- | Create a new 'PantryConfig' with the given settings.
 --
@@ -363,44 +379,63 @@ withPantryConfig root pic he count pullURL maxPerRequest =
 withPantryConfig'
   :: HasLogFunc env
   => Path Abs Dir
-  -- ^ pantry root directory, where the SQLite database and Hackage
-  -- downloads are kept.
+     -- ^ pantry root directory, where the SQLite database and Hackage
+     -- downloads are kept.
   -> PackageIndexConfig
-  -- ^ Package index configuration. You probably want
-  -- 'defaultPackageIndexConfig'.
+     -- ^ Package index configuration. You probably want
+     -- 'defaultPackageIndexConfig'.
   -> HpackExecutable
-  -- ^ When converting an hpack @package.yaml@ file to a cabal file,
-  -- what version of hpack should we use?
+     -- ^ When converting an hpack @package.yaml@ file to a cabal file,
+     -- what version of hpack should we use?
+  -> Hpack.Force
+     -- ^ Should Hpack force the overwriting of a Cabal file that has been
+     -- modified manually?
+     --
+     -- @since 0.10.0
   -> Int
-  -- ^ Maximum connection count
+     -- ^ Maximum connection count
   -> Maybe (CasaRepoPrefix, Int)
-  -- ^ Optionally, the Casa pull URL e.g. @https://casa.fpcomplete.com@ and the
-  -- maximum number of Casa keys to pull per request.
+     -- ^ Optionally, the Casa pull URL e.g. @https://casa.fpcomplete.com@ and
+     -- the maximum number of Casa keys to pull per request.
   -> (SnapName -> RawSnapshotLocation)
-  -- ^ The location of snapshot synonyms
+     -- ^ The location of snapshot synonyms
+  -> (WantedCompiler -> GlobalHintsLocation)
+     -- ^ The location of global hints
   -> (PantryConfig -> RIO env a)
-  -- ^ What to do with the config
+     -- ^ What to do with the config
   -> RIO env a
-withPantryConfig' root pic he count mCasaConfig snapLoc inner = do
-  env <- ask
-  pantryRelFile <- parseRelFile "pantry.sqlite3"
-  -- Silence persistent's logging output, which is really noisy
-  runRIO (mempty :: LogFunc) $ initStorage (root </> pantryRelFile) $ \storage -> runRIO env $ do
-    ur <- newMVar True
-    ref1 <- newIORef mempty
-    ref2 <- newIORef mempty
-    inner PantryConfig
-      { pcPackageIndex = pic
-      , pcHpackExecutable = he
-      , pcRootDir = root
-      , pcStorage = storage
-      , pcUpdateRef = ur
-      , pcConnectionCount = count
-      , pcParsedCabalFilesRawImmutable = ref1
-      , pcParsedCabalFilesMutable = ref2
-      , pcCasaConfig = mCasaConfig
-      , pcSnapshotLocation = snapLoc
-      }
+withPantryConfig'
+    root
+    pic
+    he
+    hpackForce
+    count
+    mCasaConfig
+    snapLoc
+    globalHintsLoc
+    inner
+  = do
+    env <- ask
+    pantryRelFile <- parseRelFile "pantry.sqlite3"
+    -- Silence persistent's logging output, which is really noisy
+    runRIO (mempty :: LogFunc) $ initStorage (root </> pantryRelFile) $ \storage -> runRIO env $ do
+      ur <- newMVar True
+      ref1 <- newIORef mempty
+      ref2 <- newIORef mempty
+      inner PantryConfig
+        { pcPackageIndex = pic
+        , pcHpackExecutable = he
+        , pcHpackForce = hpackForce
+        , pcRootDir = root
+        , pcStorage = storage
+        , pcUpdateRef = ur
+        , pcConnectionCount = count
+        , pcParsedCabalFilesRawImmutable = ref1
+        , pcParsedCabalFilesMutable = ref2
+        , pcCasaConfig = mCasaConfig
+        , pcSnapshotLocation = snapLoc
+        , pcGlobalHintsLocation = globalHintsLoc
+        }
 
 -- | Default pull URL for Casa.
 --
@@ -485,7 +520,8 @@ getLatestHackageRevision req name version = do
     Nothing -> pure Nothing
     Just (revision, cfKey@(BlobKey sha size)) -> do
       let cfi = CFIHash sha (Just size)
-      treeKey' <- getHackageTarballKey (PackageIdentifierRevision name version cfi)
+      treeKey' <-
+        getHackageTarballKey (PackageIdentifierRevision name version cfi)
       pure $ Just (revision, cfKey, treeKey')
 
 -- | Fetch keys and blobs and insert into the database where possible.
@@ -612,7 +648,8 @@ fetchPackages pls = do
   archives = run archivesE
   repos = run reposE
 
-  go (PLIHackage ident cfHash tree) = (s (toPir ident cfHash, Just tree), mempty, mempty)
+  go (PLIHackage ident cfHash tree) =
+    (s (toPir ident cfHash, Just tree), mempty, mempty)
   go (PLIArchive archive pm) = (mempty, s (archive, pm), mempty)
   go (PLIRepo repo pm) = (mempty, mempty, s (repo, pm))
 
@@ -658,10 +695,11 @@ loadCabalFileImmutable loc = withCache $ do
   (_warnings, gpd) <- rawParseGPD (Left $ toRawPLI loc) bs
   let pm =
         case loc of
-          PLIHackage (PackageIdentifier name version) _cfHash mtree -> PackageMetadata
-            { pmIdent = PackageIdentifier name version
-            , pmTreeKey = mtree
-            }
+          PLIHackage (PackageIdentifier name version) _cfHash mtree ->
+            PackageMetadata
+              { pmIdent = PackageIdentifier name version
+              , pmTreeKey = mtree
+              }
           PLIArchive _ pm' -> pm'
           PLIRepo _ pm' -> pm'
   let exc = MismatchedPackageMetadata (toRawPLI loc) (toRawPM pm) Nothing
@@ -701,11 +739,12 @@ loadCabalFileRawImmutable loc = withCache $ do
   (_warnings, gpd) <- rawParseGPD (Left loc) bs
   let rpm =
         case loc of
-          RPLIHackage (PackageIdentifierRevision name version _cfi) mtree -> RawPackageMetadata
-            { rpmName = Just name
-            , rpmVersion = Just version
-            , rpmTreeKey = mtree
-            }
+          RPLIHackage (PackageIdentifierRevision name version _cfi) mtree ->
+            RawPackageMetadata
+              { rpmName = Just name
+              , rpmVersion = Just version
+              , rpmTreeKey = mtree
+              }
           RPLIArchive _ rpm' -> rpm'
           RPLIRepo _ rpm' -> rpm'
   let exc = MismatchedPackageMetadata loc rpm Nothing (gpdPackageIdentifier gpd)
@@ -780,41 +819,41 @@ loadCabalFilePath progName dir = do
       let gpdio = run . getGPD cabalfp gpdRef
           triple = (gpdio, name, cabalfp)
       atomicModifyIORef' ref $ \m -> (Map.insert dir triple m, triple)
-  where
-    getGPD cabalfp gpdRef printWarnings = do
-      mpair <- readIORef gpdRef
-      (warnings0, gpd) <-
-        case mpair of
-          Just pair -> pure pair
-          Nothing -> do
-            bs <- liftIO $ B.readFile $ toFilePath cabalfp
-            (warnings0, gpd) <- rawParseGPD (Right cabalfp) bs
-            checkCabalFileName (gpdPackageName gpd) cabalfp
-            pure (warnings0, gpd)
-      warnings <-
-        case printWarnings of
-          YesPrintWarnings -> mapM_ (logWarn . toPretty cabalfp) warnings0 $> []
-          NoPrintWarnings -> pure warnings0
-      writeIORef gpdRef $ Just (warnings, gpd)
-      pure gpd
+ where
+  getGPD cabalfp gpdRef printWarnings = do
+    mpair <- readIORef gpdRef
+    (warnings0, gpd) <-
+      case mpair of
+        Just pair -> pure pair
+        Nothing -> do
+          bs <- liftIO $ B.readFile $ toFilePath cabalfp
+          (warnings0, gpd) <- rawParseGPD (Right cabalfp) bs
+          checkCabalFileName (gpdPackageName gpd) cabalfp
+          pure (warnings0, gpd)
+    warnings <-
+      case printWarnings of
+        YesPrintWarnings -> mapM_ (logWarn . toPretty cabalfp) warnings0 $> []
+        NoPrintWarnings -> pure warnings0
+    writeIORef gpdRef $ Just (warnings, gpd)
+    pure gpd
 
-    toPretty :: Path Abs File -> PWarning -> Utf8Builder
-    toPretty src (PWarning _type pos msg) =
-      "Cabal file warning in " <>
-      fromString (toFilePath src) <> "@" <>
-      fromString (showPos pos) <> ": " <>
-      fromString msg
+  toPretty :: Path Abs File -> PWarning -> Utf8Builder
+  toPretty src (PWarning _type pos msg) =
+    "Cabal file warning in " <>
+    fromString (toFilePath src) <> "@" <>
+    fromString (showPos pos) <> ": " <>
+    fromString msg
 
-    -- | Check if the given name in the @Package@ matches the name of the .cabal
-    -- file
-    checkCabalFileName :: MonadThrow m => PackageName -> Path Abs File -> m ()
-    checkCabalFileName name cabalfp = do
-      -- Previously, we just use parsePackageNameFromFilePath. However, that can
-      -- lead to confusing error messages. See:
-      -- https://github.com/commercialhaskell/stack/issues/895
-      let expected = T.unpack $ unSafeFilePath $ cabalFileName name
-      when (expected /= toFilePath (filename cabalfp)) $
-        throwM $ MismatchedCabalName cabalfp name
+  -- | Check if the given name in the @Package@ matches the name of the .cabal
+  -- file
+  checkCabalFileName :: MonadThrow m => PackageName -> Path Abs File -> m ()
+  checkCabalFileName name cabalfp = do
+    -- Previously, we just use parsePackageNameFromFilePath. However, that can
+    -- lead to confusing error messages. See:
+    -- https://github.com/commercialhaskell/stack/issues/895
+    let expected = T.unpack $ unSafeFilePath $ cabalFileName name
+    when (expected /= toFilePath (filename cabalfp)) $
+      throwM $ MismatchedCabalName cabalfp name
 
 -- | Get the file name for the Cabal file in the given directory.
 --
@@ -870,17 +909,19 @@ hpack progName pkgDir = do
   when exists $ do
     logDebug $ "Running Hpack on " <> fromString (toFilePath hpackFile)
     he <- view $ pantryConfigL.to pcHpackExecutable
+    hpackForce <- view $ pantryConfigL.to pcHpackForce
     case he of
       HpackBundled ->
-                  liftIO
-                     ( Hpack.hpackResultWithError
-                     $ mHpackProgName
-                     $ Hpack.setDecode decodeYaml
-                     $ Hpack.setFormatYamlParseError formatYamlParseError
-                     $ Hpack.setTarget
-                         (toFilePath hpackFile) Hpack.defaultOptions
-                     )
-                   >>= \ case
+        liftIO
+           ( Hpack.hpackResultWithError
+           $ mHpackProgName
+           $ Hpack.setDecode decodeYaml
+           $ Hpack.setFormatYamlParseError formatYamlParseError
+           $ Hpack.setTarget
+               (toFilePath hpackFile)
+               Hpack.defaultOptions { Hpack.optionsForce = hpackForce }
+           )
+         >>= \ case
         Left err -> throwIO (HpackLibraryException hpackFile $ formatHpackError (fromMaybe "hpack" progName) err)
         Right r -> do
           forM_ (Hpack.resultWarnings r) (logWarn . fromString)
@@ -909,11 +950,15 @@ hpack progName pkgDir = do
               <> fromString (toFilePath (filename hpackFile))
               <> " file instead of the Cabal file,\n"
               <> "then please delete the Cabal file."
-      HpackCommand command -> catchAny
-        ( withWorkingDir (toFilePath pkgDir) $
-          proc command [] runProcess_
-        )
-        ( throwIO . HpackExeException command pkgDir)
+      HpackCommand command -> do
+        let hpackArgs = case hpackForce of
+              Hpack.Force -> ["--force"]
+              Hpack.NoForce -> []
+        catchAny
+          ( withWorkingDir (toFilePath pkgDir) $
+            proc command hpackArgs runProcess_
+          )
+          ( throwIO . HpackExeException command pkgDir)
 
 -- | Get the 'PackageIdentifier' from a 'GenericPackageDescription'.
 --
@@ -1448,10 +1493,16 @@ warnUnusedAddPackagesConfig ::
   => Utf8Builder -- ^ source
   -> AddPackagesConfig
   -> RIO env ()
-warnUnusedAddPackagesConfig source (AddPackagesConfig _drops flags hiddens options) = do
-  unless (null ls) $ do
-    logWarn $ "Some warnings discovered when adding packages to snapshot (" <> source <> ")"
-    traverse_ logWarn ls
+warnUnusedAddPackagesConfig
+    source
+    (AddPackagesConfig _drops flags hiddens options)
+  = do
+    unless (null ls) $ do
+      logWarn $
+           "Some warnings discovered when adding packages to snapshot ("
+        <> source
+        <> ")"
+      traverse_ logWarn ls
  where
   ls = concat [flags', hiddens', options']
 
@@ -1495,39 +1546,46 @@ addPackagesToSnapshot ::
   -> AddPackagesConfig
   -> Map PackageName RawSnapshotPackage -- ^ packages from parent
   -> RIO env (Map PackageName RawSnapshotPackage, AddPackagesConfig)
-addPackagesToSnapshot source newPackages (AddPackagesConfig drops flags hiddens options) old = do
-  new' <- for newPackages $ \loc -> do
-    name <- getPackageLocationName loc
-    pure (name, RawSnapshotPackage
-      { rspLocation = loc
-      , rspFlags = Map.findWithDefault mempty name flags
-      , rspHidden = Map.findWithDefault False name hiddens
-      , rspGhcOptions = Map.findWithDefault [] name options
-      })
-  let (newSingles, newMultiples)
-        = partitionEithers
-        $ map sonToEither
-        $ Map.toList
-        $ Map.fromListWith (<>)
-        $ map (second Single) new'
-  unless (null newMultiples) $ throwIO $
-    DuplicatePackageNames source $ map (second (map rspLocation)) newMultiples
-  let new = Map.fromList newSingles
-      allPackages0 = new `Map.union` (old `Map.difference` Map.fromSet (const ()) drops)
-      allPackages = flip Map.mapWithKey allPackages0 $ \name rsp ->
-        rsp
-          { rspFlags = Map.findWithDefault (rspFlags rsp) name flags
-          , rspHidden = Map.findWithDefault (rspHidden rsp) name hiddens
-          , rspGhcOptions = Map.findWithDefault (rspGhcOptions rsp) name options
-          }
+addPackagesToSnapshot
+    source
+    newPackages
+    (AddPackagesConfig drops flags hiddens options)
+    old
+  = do
+    new' <- for newPackages $ \loc -> do
+      name <- getPackageLocationName loc
+      pure (name, RawSnapshotPackage
+        { rspLocation = loc
+        , rspFlags = Map.findWithDefault mempty name flags
+        , rspHidden = Map.findWithDefault False name hiddens
+        , rspGhcOptions = Map.findWithDefault [] name options
+        })
+    let (newSingles, newMultiples)
+          = partitionEithers
+          $ map sonToEither
+          $ Map.toList
+          $ Map.fromListWith (<>)
+          $ map (second Single) new'
+    unless (null newMultiples) $ throwIO $
+      DuplicatePackageNames source $ map (second (map rspLocation)) newMultiples
+    let new = Map.fromList newSingles
+        allPackages0 =
+          new `Map.union` (old `Map.difference` Map.fromSet (const ()) drops)
+        allPackages = flip Map.mapWithKey allPackages0 $ \name rsp ->
+          rsp
+            { rspFlags = Map.findWithDefault (rspFlags rsp) name flags
+            , rspHidden = Map.findWithDefault (rspHidden rsp) name hiddens
+            , rspGhcOptions =
+                Map.findWithDefault (rspGhcOptions rsp) name options
+            }
 
-      unused = AddPackagesConfig
-        (drops `Set.difference` Map.keysSet old)
-        (flags `Map.difference` allPackages)
-        (hiddens `Map.difference` allPackages)
-        (options `Map.difference` allPackages)
+        unused = AddPackagesConfig
+          (drops `Set.difference` Map.keysSet old)
+          (flags `Map.difference` allPackages)
+          (hiddens `Map.difference` allPackages)
+          (options `Map.difference` allPackages)
 
-  pure (allPackages, unused)
+    pure (allPackages, unused)
 
 cachedSnapshotCompletePackageLocation ::
      (HasPantryConfig env, HasLogFunc env, HasProcessContext env)
@@ -1566,56 +1624,62 @@ addAndCompletePackagesToSnapshot ::
   -> RIO
        env
        (Map PackageName SnapshotPackage, [CompletedPLI], AddPackagesConfig)
-addAndCompletePackagesToSnapshot loc cachedPL newPackages (AddPackagesConfig drops flags hiddens options) old = do
-  let source = display loc
-      addPackage ::
-           (HasPantryConfig env, HasLogFunc env, HasProcessContext env)
-        => ([(PackageName, SnapshotPackage)],[CompletedPLI])
-        -> RawPackageLocationImmutable
-        -> RIO env ([(PackageName, SnapshotPackage)], [CompletedPLI])
-      addPackage (ps, completed) rawLoc = do
-        mcomplLoc <- cachedSnapshotCompletePackageLocation cachedPL rawLoc
-        case mcomplLoc of
-          Nothing -> do
-            warnMissingCabalFile rawLoc
-            pure (ps, completed)
-          Just complLoc -> do
-            let PackageIdentifier name _ = packageLocationIdent complLoc
-                p = (name, SnapshotPackage
-                  { spLocation = complLoc
-                  , spFlags = Map.findWithDefault mempty name flags
-                  , spHidden = Map.findWithDefault False name hiddens
-                  , spGhcOptions = Map.findWithDefault [] name options
-                  })
-                completed' = if toRawPLI complLoc == rawLoc
-                             then completed
-                             else CompletedPLI rawLoc complLoc:completed
-            pure (p:ps, completed')
-  (revNew, revCompleted) <- foldM addPackage ([], []) newPackages
-  let (newSingles, newMultiples)
-        = partitionEithers
-        $ map sonToEither
-        $ Map.toList
-        $ Map.fromListWith (<>)
-        $ map (second Single) (reverse revNew)
-  unless (null newMultiples) $ throwIO $
-    DuplicatePackageNames source $ map (second (map (toRawPLI . spLocation))) newMultiples
-  let new = Map.fromList newSingles
-      allPackages0 = new `Map.union` (old `Map.difference` Map.fromSet (const ()) drops)
-      allPackages = flip Map.mapWithKey allPackages0 $ \name sp ->
-        sp
-          { spFlags = Map.findWithDefault (spFlags sp) name flags
-          , spHidden = Map.findWithDefault (spHidden sp) name hiddens
-          , spGhcOptions = Map.findWithDefault (spGhcOptions sp) name options
-          }
+addAndCompletePackagesToSnapshot
+    loc
+    cachedPL
+    newPackages
+    (AddPackagesConfig drops flags hiddens options)
+    old
+  = do
+    let source = display loc
+        addPackage ::
+             (HasPantryConfig env, HasLogFunc env, HasProcessContext env)
+          => ([(PackageName, SnapshotPackage)],[CompletedPLI])
+          -> RawPackageLocationImmutable
+          -> RIO env ([(PackageName, SnapshotPackage)], [CompletedPLI])
+        addPackage (ps, completed) rawLoc = do
+          mcomplLoc <- cachedSnapshotCompletePackageLocation cachedPL rawLoc
+          case mcomplLoc of
+            Nothing -> do
+              warnMissingCabalFile rawLoc
+              pure (ps, completed)
+            Just complLoc -> do
+              let PackageIdentifier name _ = packageLocationIdent complLoc
+                  p = (name, SnapshotPackage
+                    { spLocation = complLoc
+                    , spFlags = Map.findWithDefault mempty name flags
+                    , spHidden = Map.findWithDefault False name hiddens
+                    , spGhcOptions = Map.findWithDefault [] name options
+                    })
+                  completed' = if toRawPLI complLoc == rawLoc
+                               then completed
+                               else CompletedPLI rawLoc complLoc:completed
+              pure (p:ps, completed')
+    (revNew, revCompleted) <- foldM addPackage ([], []) newPackages
+    let (newSingles, newMultiples)
+          = partitionEithers
+          $ map sonToEither
+          $ Map.toList
+          $ Map.fromListWith (<>)
+          $ map (second Single) (reverse revNew)
+    unless (null newMultiples) $ throwIO $
+      DuplicatePackageNames source $ map (second (map (toRawPLI . spLocation))) newMultiples
+    let new = Map.fromList newSingles
+        allPackages0 = new `Map.union` (old `Map.difference` Map.fromSet (const ()) drops)
+        allPackages = flip Map.mapWithKey allPackages0 $ \name sp ->
+          sp
+            { spFlags = Map.findWithDefault (spFlags sp) name flags
+            , spHidden = Map.findWithDefault (spHidden sp) name hiddens
+            , spGhcOptions = Map.findWithDefault (spGhcOptions sp) name options
+            }
 
-      unused = AddPackagesConfig
-        (drops `Set.difference` Map.keysSet old)
-        (flags `Map.difference` allPackages)
-        (hiddens `Map.difference` allPackages)
-        (options `Map.difference` allPackages)
+        unused = AddPackagesConfig
+          (drops `Set.difference` Map.keysSet old)
+          (flags `Map.difference` allPackages)
+          (hiddens `Map.difference` allPackages)
+          (options `Map.difference` allPackages)
 
-  pure (allPackages, reverse revCompleted, unused)
+    pure (allPackages, reverse revCompleted, unused)
 
 -- | Parse a 'SnapshotLayer' value from a 'SnapshotLocation'.
 --
@@ -1638,7 +1702,8 @@ loadRawSnapshotLayer rsl@(RSLUrl url blob) =
 loadRawSnapshotLayer rsl@(RSLFilePath fp) =
   handleAny (throwIO . InvalidSnapshot rsl) $ do
     value <- Yaml.decodeFileThrow $ toFilePath $ resolvedAbsolute fp
-    snapshot <- warningsParserHelperRaw rsl value $ Just $ parent $ resolvedAbsolute fp
+    snapshot <-
+      warningsParserHelperRaw rsl value $ Just $ parent $ resolvedAbsolute fp
     pure $ Right (snapshot, CompletedSL rsl (SLFilePath fp))
 loadRawSnapshotLayer rsl@(RSLSynonym syn) = do
   loc <- snapshotLocation syn
@@ -1668,7 +1733,8 @@ loadSnapshotLayer sl@(SLUrl url blob) =
 loadSnapshotLayer sl@(SLFilePath fp) =
   handleAny (throwIO . InvalidSnapshot (toRawSL sl)) $ do
     value <- Yaml.decodeFileThrow $ toFilePath $ resolvedAbsolute fp
-    snapshot <- warningsParserHelper sl value $ Just $ parent $ resolvedAbsolute fp
+    snapshot <-
+      warningsParserHelper sl value $ Just $ parent $ resolvedAbsolute fp
     pure $ Right snapshot
 
 loadFromURL ::
@@ -1699,8 +1765,11 @@ loadUrlViaCasaOrWithCheck url blobKey = do
   case mblobFromCasa of
     Just blob -> do
       logDebug
-        ("Loaded snapshot from Casa (" <> display blobKey <> ") for URL: " <>
-         display url)
+        (  "Loaded snapshot from Casa ("
+        <> display blobKey
+        <> ") for URL: "
+        <> display url
+        )
       pure blob
     Nothing -> loadWithCheck url (Just blobKey)
 
@@ -1786,19 +1855,22 @@ getRawPackageLocationIdent ::
      (HasPantryConfig env, HasLogFunc env, HasProcessContext env)
   => RawPackageLocationImmutable
   -> RIO env PackageIdentifier
-getRawPackageLocationIdent (RPLIHackage (PackageIdentifierRevision name version _) _) =
-  pure $ PackageIdentifier name version
-getRawPackageLocationIdent (RPLIRepo _ RawPackageMetadata { rpmName = Just name, rpmVersion = Just version }) =
-  pure $ PackageIdentifier name version
-getRawPackageLocationIdent (RPLIArchive _ RawPackageMetadata { rpmName = Just name, rpmVersion = Just version }) =
-  pure $ PackageIdentifier name version
+getRawPackageLocationIdent
+  (RPLIHackage (PackageIdentifierRevision name version _) _) =
+    pure $ PackageIdentifier name version
+getRawPackageLocationIdent
+  (RPLIRepo _ RawPackageMetadata { rpmName = Just name, rpmVersion = Just version }) =
+    pure $ PackageIdentifier name version
+getRawPackageLocationIdent
+  (RPLIArchive _ RawPackageMetadata { rpmName = Just name, rpmVersion = Just version }) =
+    pure $ PackageIdentifier name version
 getRawPackageLocationIdent rpli = packageIdent <$> loadPackageRaw rpli
 
 -- | Get the 'TreeKey' of the package at the given location.
 --
 -- @since 0.1.0.0
-getRawPackageLocationTreeKey
-  :: (HasPantryConfig env, HasLogFunc env, HasProcessContext env)
+getRawPackageLocationTreeKey ::
+     (HasPantryConfig env, HasLogFunc env, HasProcessContext env)
   => RawPackageLocationImmutable
   -> RIO env TreeKey
 getRawPackageLocationTreeKey pl =
@@ -1813,8 +1885,8 @@ getRawPackageLocationTreeKey pl =
 -- | Get the 'TreeKey' of the package at the given location.
 --
 -- @since 0.1.0.0
-getPackageLocationTreeKey
-  :: (HasPantryConfig env, HasLogFunc env, HasProcessContext env)
+getPackageLocationTreeKey ::
+     (HasPantryConfig env, HasLogFunc env, HasProcessContext env)
   => PackageLocationImmutable
   -> RIO env TreeKey
 getPackageLocationTreeKey pl = pure $ getTreeKey pl
@@ -1871,12 +1943,23 @@ data PantryApp = PantryApp
 simpleAppL :: Lens' PantryApp SimpleApp
 simpleAppL = lens paSimpleApp (\x y -> x { paSimpleApp = y })
 
--- | Lens to view or modify the 'HpackExecutable' of a 'PantryConfig'
+-- | Lens to view or modify the 'HpackExecutable' of a 'PantryConfig'.
 --
 -- @since 0.1.0.0
 hpackExecutableL :: Lens' PantryConfig HpackExecutable
 hpackExecutableL k pconfig =
-  fmap (\hpExe -> pconfig { pcHpackExecutable = hpExe }) (k (pcHpackExecutable pconfig))
+  fmap
+    (\hpExe -> pconfig { pcHpackExecutable = hpExe })
+    (k (pcHpackExecutable pconfig))
+
+-- | Lens to view or modify the 'Hpack.Force' of a 'PantryConfig'.
+--
+-- @since 0.10.0
+hpackForceL :: Lens' PantryConfig Hpack.Force
+hpackForceL k pconfig =
+  fmap
+    (\hpackForce -> pconfig { pcHpackForce = hpackForce })
+    (k (pcHpackForce pconfig))
 
 instance HasLogFunc PantryApp where
   logFuncL = simpleAppL.logFuncL
@@ -1914,27 +1997,34 @@ runPantryAppWith ::
   -> Int
   -> RIO PantryApp a
   -> m a
-runPantryAppWith maxConnCount casaRepoPrefix casaMaxPerRequest f = runSimpleApp $ do
-  sa <- ask
-  stack <- getAppUserDataDirectory "stack"
-  root <- parseAbsDir $ stack FilePath.</> "pantry"
-  withPantryConfig'
-    root
-    defaultPackageIndexConfig
-    HpackBundled
+runPantryAppWith
     maxConnCount
-    (Just (casaRepoPrefix, casaMaxPerRequest))
-    defaultSnapshotLocation
-    $ \pc ->
-      runRIO
-        PantryApp
-          { paSimpleApp = sa
-          , paPantryConfig = pc
-          , paTermWidth = 100
-          , paUseColor = True
-          , paStylesUpdate = mempty
-          }
-        f
+    casaRepoPrefix
+    casaMaxPerRequest
+    f
+  = runSimpleApp $ do
+    sa <- ask
+    stack <- getAppUserDataDirectory "stack"
+    root <- parseAbsDir $ stack FilePath.</> "pantry"
+    withPantryConfig'
+      root
+      defaultPackageIndexConfig
+      HpackBundled
+      Hpack.NoForce
+      maxConnCount
+      (Just (casaRepoPrefix, casaMaxPerRequest))
+      defaultSnapshotLocation
+      defaultGlobalHintsLocation
+      $ \pc ->
+        runRIO
+          PantryApp
+            { paSimpleApp = sa
+            , paPantryConfig = pc
+            , paTermWidth = 100
+            , paUseColor = True
+            , paStylesUpdate = mempty
+            }
+          f
 
 -- | Like 'runPantryApp', but uses an empty pantry directory instead of sharing
 -- with Stack. Useful for testing.
@@ -1949,9 +2039,11 @@ runPantryAppClean f =
       root
       defaultPackageIndexConfig
       HpackBundled
+      Hpack.NoForce
       8
       (Just (defaultCasaRepoPrefix, defaultCasaMaxPerRequest))
       defaultSnapshotLocation
+      defaultGlobalHintsLocation
       $ \pc ->
         runRIO
           PantryApp
@@ -1963,46 +2055,72 @@ runPantryAppClean f =
             }
           f
 
--- | Load the global hints from GitHub.
+-- | Load the global hints.
 --
--- @since 0.1.0.0
+-- @since 9.4.0
 loadGlobalHints ::
      (HasTerm env, HasPantryConfig env)
   => WantedCompiler
   -> RIO env (Maybe (Map PackageName Version))
-loadGlobalHints wc =
-    inner False
+loadGlobalHints wc = do
+  dest <- getGlobalHintsFile
+  loc <- globalHintsLocation wc
+  inner dest loc False
  where
-  inner alreadyDownloaded = do
-    dest <- getGlobalHintsFile
-    req <- parseRequest "https://raw.githubusercontent.com/commercialhaskell/stackage-content/master/stack/global-hints.yaml"
-    downloaded <- download req dest
-    eres <- tryAny (inner2 dest)
-    mres <-
-      case eres of
-        Left e -> Nothing <$ logError
-                               ( "Error: [S-912]\n"
-                                 <> "Error when parsing global hints: "
-                                 <> displayShow e
-                               )
-        Right x -> pure x
-    case mres of
-      Nothing | not alreadyDownloaded && not downloaded -> do
-        logInfo $
-          "Could not find local global hints for " <>
-          RIO.display wc <>
-          ", forcing a redownload"
-        x <- redownload req dest
-        if x
-          then inner True
-          else do
-            logInfo "Redownload didn't happen"
-            pure Nothing
-      _ -> pure mres
-
-  inner2 dest = liftIO $
-        Map.lookup wc . fmap (fmap unCabalString . unCabalStringMap)
-    <$> Yaml.decodeFileThrow (toFilePath dest)
+  inner dest loc alreadyDownloaded = case loc of
+    GHLUrl url -> do
+      req <- parseRequest $ T.unpack url
+      downloaded <- download req dest
+      mres <- tryParseYaml dest
+      case mres of
+        Nothing | not alreadyDownloaded && not downloaded -> do
+          prettyInfoL
+            [ flow "Could not find local global hints for"
+            , string (T.unpack $ RIO.textDisplay wc) <> ","
+            , flow "forcing a redownload."
+            ]
+          redownloaded <- redownload req dest
+          if redownloaded
+            then inner dest loc True
+            else do
+              logInfo "Redownload didn't happen"
+              pure Nothing
+        _ -> pure mres
+    GHLFilePath fp -> do
+      let source = resolvedAbsolute fp
+      mres <- tryParseYaml source
+      case mres of
+        Nothing -> do
+          prettyInfoL
+            [ flow "Could not find local global hints for"
+            , string (T.unpack $ RIO.textDisplay wc)
+            , "in"
+            , pretty source <> "."
+            ]
+          pure Nothing
+        _ -> do
+          liftIO $ copyFile source dest
+          prettyDebugL
+            [ flow "Installed global hints from"
+            , pretty source
+            ]
+          pure mres
+  inner2 fp = liftIO $ do
+    allGlobalHints <- Yaml.decodeFileThrow (toFilePath fp)
+    let globalHints = Map.lookup wc allGlobalHints
+    pure $ fmap (fmap unCabalString . unCabalStringMap) globalHints
+  tryParseYaml fp = do
+    eres <- tryAny (inner2 fp)
+    case eres of
+      Left e -> do
+        prettyError $
+          "[S-912]"
+          <> line
+          <> flow "Error when parsing global hints:"
+          <> blankLine
+          <> string (displayException e)
+        pure Nothing
+      Right x -> pure x
 
 -- | Partition a map of global packages with its versions into a Set of replaced
 -- packages and its dependencies and a map of remaining (untouched) packages.
